import filecmp
import logging
import time
from datetime import datetime, timedelta
from pathlib import Path
from threading import Event
from typing import Any, Generator, List, Type
from unittest.mock import MagicMock, _Call, call, patch

import pytest
import pyvisa
from pytest import LogCaptureFixture

# pyright thinks this constant is not exported
from pyvisa.errors import VI_ERROR_TMO  # type: ignore
from pyvisa.resources import RegisterBasedResource, SerialInstrument

from psc_datalogger.connection import (
    ConnectionManager,
    ConnectionNotInitialized,
    DataWriter,
    InstrumentConfig,
    PrologixNotFoundException,
    Worker,
)
from psc_datalogger.multimeter import Agilent3458A, Agilent34401A, Multimeter


class TestConnectionManager:
    """Testing the ConnectionManager class"""

    @pytest.fixture
    def connmgr(self, qtbot) -> Generator[ConnectionManager, Any, Any]:
        connmgr = ConnectionManager()
        self.mock_status_bar = MagicMock()
        connmgr.set_status_bar(self.mock_status_bar)
        yield connmgr

        if connmgr.thread.isRunning():
            connmgr.thread.quit()
            connmgr._worker._exit()
            time.sleep(0.1)
            qtbot.waitSignal(connmgr.thread.finished)
            assert connmgr.thread.isFinished()

    @patch.object(Worker, "init_connection")
    def test_start(self, mock_init_connection, connmgr: ConnectionManager, qtbot):
        """Test that calling the start() function triggers the thread to begin"""
        connmgr.start()

        qtbot.waitUntil(lambda: connmgr.thread.isRunning())

        # Time for Worker to begin executing its run() method
        time.sleep(0.1)
        mock_init_connection.assert_called_once()

    def test_register_status_bar(
        self,
    ):
        """Test that the ConnectionManager registers the callbacks into the StatusBar
        correctly.

        NOT using connmgr fixture - need to test the behaviour it sets up"""
        mock_status_bar = MagicMock()
        conn_manager = ConnectionManager()
        conn_manager.set_status_bar(mock_status_bar)

        now = datetime.now()
        errmsg = "Test error message"
        # Invoke each of the callbacks
        conn_manager._worker.query_complete.emit(now)
        conn_manager._worker.init_complete.emit()
        conn_manager._worker.error.emit(errmsg)

        # Confirm callbacks triggered
        mock_status_bar.query_complete_callback.assert_called_once_with(now)
        mock_status_bar.init_complete_callback.assert_called_once()
        mock_status_bar.error_callback.assert_called_once_with(errmsg)

    @patch.object(Worker, "validate_parameters")
    def test_start_logging_valid_params(
        self, mock_valid_parameters: MagicMock, connmgr: ConnectionManager
    ):
        """Test that if parameters are valid, the correct actions will occur"""
        mock_valid_parameters.return_value = True

        assert connmgr.start_logging() is True

        assert connmgr.logging_signal.is_set()
        self.mock_status_bar.logging_started.assert_called_once()

    @patch.object(Worker, "validate_parameters")
    def test_start_logging_invalid_params(
        self, mock_valid_parameters: MagicMock, connmgr: ConnectionManager
    ):
        """Test that if parameters are invalid, the correct actions will occur"""
        mock_valid_parameters.return_value = False

        assert connmgr.start_logging() is False

        assert connmgr.logging_signal.is_set() is False
        self.mock_status_bar.logging_started.assert_not_called()

    def test_stop_logging(self, connmgr: ConnectionManager):
        """Test that stop_logging sends the expected signals"""
        connmgr.stop_logging()

        assert connmgr.logging_signal.is_set() is False
        self.mock_status_bar.logging_stopped.assert_called_once()

    def test_set_instrument_handles_expected_exception(
        self, connmgr: ConnectionManager, caplog: pytest.LogCaptureFixture
    ):
        """Test that set_instrument handles the ConnectionNotInitialized exception"""

        mocked_worker = MagicMock()
        mock_set_instrument = MagicMock(side_effect=ConnectionNotInitialized)
        mocked_worker.set_instrument = mock_set_instrument
        connmgr._worker = mocked_worker

        with caplog.at_level(logging.ERROR):
            connmgr.set_instrument(
                0, False, "10", False, Agilent3458A
            )  # parameters don't matter

        # Check that the expected message appears
        assert "Could not set instrument" in caplog.text

    def test_set_nplc_handles_expected_exception(
        self, connmgr: ConnectionManager, caplog: pytest.LogCaptureFixture
    ):
        """Test that set_nplc handles the ConnectionNotInitialized exception"""

        mocked_worker = MagicMock()
        mock_set_nplc = MagicMock(side_effect=ConnectionNotInitialized)
        mocked_worker.set_nplc = mock_set_nplc
        connmgr._worker = mocked_worker

        with caplog.at_level(logging.ERROR):
            connmgr.set_nplc("123")  # parameters don't matter

        # Check that the expected message appears
        assert "Could not set NPLC" in caplog.text


class TestDataWriter:
    """Testing the DataWriter class"""

    filepath: Path

    @pytest.fixture
    def datawriter(self, tmp_path: Path) -> Generator[DataWriter, Any, Any]:
        self.filepath = tmp_path / "test.csv"
        writer = DataWriter(str(self.filepath))
        yield writer
        writer.close()

    def test_write(self, datawriter: DataWriter):
        """Test that write correctly puts the data into the file"""
        time = datetime(2024, 3, 6, 9, 15, 30)
        datawriter.write(time, "1.23", "4.56", "7.89")

        time = time + timedelta(seconds=30)
        datawriter.write(time, "0.12", "0.34", "0.56")

        expected = Path(__file__).parent / "test_output.csv"
        # Useful debugging code
        # with self.filepath.open("r") as f:
        #     print(f.readlines())
        # with expected.open("r") as f:
        #     print(f.readlines())

        assert filecmp.cmp(self.filepath, expected)


def multimeter_naming(val):
    """Function to provide readable names for Pytest parameterized tests"""
    if isinstance(val, Multimeter):
        return val.__class__
    return ""


class TestWorker:
    logging_signal: Event

    @pytest.fixture
    def worker(self) -> Generator[Worker, Any, Any]:
        self.logging_signal = Event()
        worker = Worker(self.logging_signal)
        yield worker
        worker._exit()

    def test_set_filepath(self, worker: Worker, tmp_path: Path):
        """Test that setting a filepath creates a writer"""
        file = tmp_path / "something.csv"
        worker.set_filepath(file.as_posix())

        assert worker.writer is not None

    def test_set_filepath_repeatedly(self, worker: Worker, tmp_path: Path):
        """Test that repeatedly setting filepaths creates a new writer each time"""
        file = tmp_path / "something.csv"
        worker.set_filepath(file.as_posix())

        assert worker.writer is not None
        old_writer = worker.writer

        file = tmp_path / "somethingelse.csv"
        worker.set_filepath(file.as_posix())
        assert worker.writer != old_writer

    def test_set_instrument(self, worker: Worker):
        """Test that setting an instrument saves it"""
        # Mock out _init_instrument as it does hardware communication
        mocked_init_instrument = MagicMock()
        worker._init_instrument = mocked_init_instrument

        num = 1
        enabled = True
        address = 22
        measure_temp = False
        multimeter = MagicMock()
        multimeter_returner = MagicMock(return_value=multimeter, spec=Agilent3458A)
        worker.set_instrument(
            num,
            enabled,
            str(address),
            measure_temp,
            multimeter_returner,  # type: ignore
        )

        expected_config = InstrumentConfig(enabled, address, measure_temp, multimeter)

        multimeter_returner.assert_called_once_with()
        assert worker.instrument_configs[num] == expected_config
        mocked_init_instrument.assert_called_once_with(expected_config)

    def test_override_instrument(self, worker: Worker):
        """Test that overriding an instrument correctly replaces the config"""
        # Mock out _init_instrument as it does hardware communication
        mocked_init_instrument = MagicMock()
        worker._init_instrument = mocked_init_instrument

        # Initial values
        num = 1
        enabled = True
        address = 22
        measure_temp = False
        worker.set_instrument(num, enabled, str(address), measure_temp, Agilent3458A)

        # Override values
        enabled = False
        address = 44
        measure_temp = True
        multimeter = MagicMock()
        multimeter_returner = MagicMock(return_value=multimeter, spec=Agilent3458A)
        expected_config = InstrumentConfig(enabled, address, measure_temp, multimeter)
        worker.set_instrument(
            num,
            enabled,
            str(address),
            measure_temp,
            multimeter_returner,  # type: ignore
        )

        multimeter_returner.assert_called_once_with()
        assert worker.instrument_configs[num] == expected_config
        mocked_init_instrument.assert_called_with(expected_config)

    @pytest.mark.parametrize("invalid_value", [-1, 0, 4, 5])
    def test_set_instrument_invalid_values(self, invalid_value: int, worker: Worker):
        """Test that passing invalid values raises expected exception"""
        with pytest.raises(AssertionError):
            worker.set_instrument(invalid_value, True, "123", False, Agilent3458A)

    @pytest.mark.parametrize(
        "multimeter, expected_calls",
        [
            (
                Agilent3458A(),
                [
                    call("++addr 22"),
                    call("++auto 1"),
                    call("PRESET NORM"),
                    call("BEEP 0"),
                    call("TRIG HOLD"),
                    call("++addr 22"),
                    call("++auto 1"),
                    call("NPLC 50"),
                ],
            ),
            (
                Agilent34401A(),
                [
                    call("++addr 22"),
                    call("++auto 0"),
                    call("*RST"),
                    call("CONFigure:VOLTage:DC 10, 0.003"),
                    call("SYSTEM:BEEPER:STATE OFF"),
                    call("++addr 22"),
                    call("++auto 0"),
                    call("VOLT:DC:NPLCYCLES 50"),
                ],
            ),
        ],
        ids=multimeter_naming,
    )
    def test_init_instrument(
        self, worker: Worker, multimeter: Multimeter, expected_calls: List[_Call]
    ):
        """Test that init_instrument sends the expected calls to the hardware"""

        mocked_connection = MagicMock(spec=SerialInstrument)
        mocked_connection.bytes_in_buffer = 0
        worker._connection = mocked_connection
        worker.nplc_input = MagicMock()
        worker.nplc_input.text = MagicMock(return_value=50)

        enabled = True
        address = 22  # Note: Duplicated in the parameterized calls
        worker._init_instrument(
            InstrumentConfig(enabled, address, multimeter=multimeter)
        )

        mocked_connection.write.assert_has_calls(expected_calls, any_order=False)

    def test_init_instrument_empties_read_buffer(self, worker: Worker):
        """Test that _init_instrument reads all of the bytes out of the connection
        before returning"""
        mocked_connection_bytes_in_buffer = MagicMock(side_effect=[100, 50, 25, 0])
        worker._connection_bytes_in_buffer = mocked_connection_bytes_in_buffer

        mocked_connection = MagicMock()
        mocked_connection.read = MagicMock(side_effect=["some", "random", "data"])
        worker._connection = mocked_connection
        worker.nplc_input = MagicMock()

        enabled = True
        address = 22
        worker._init_instrument(InstrumentConfig(enabled, address))

        assert mocked_connection.read.call_count == 3

    def test_init_instrument_disabled_instrument(
        self,
        worker: Worker,
        caplog: pytest.LogCaptureFixture,
    ):
        """Check that passing a disabled instrument causes nothing to happen"""

        instrument = InstrumentConfig(False)

        with caplog.at_level(logging.DEBUG):
            worker._init_instrument(instrument)

        # Check the right log messages were created
        assert len(caplog.records) == 2
        assert (
            caplog.records[1].message
            == f"Skipping _init_instrument for instrument {instrument.address}"
        )

    @pytest.mark.parametrize("invalid_value", [0, -1])
    def test_init_instrument_invalid_values(self, invalid_value: int, worker: Worker):
        """Check that various invalid parameters raise the expected error"""
        with pytest.raises(ValueError):
            worker._init_instrument(InstrumentConfig(True, invalid_value))

    def test_set_nplc(self, worker: Worker):
        """Test that set_nplc works as expected for good values"""

        # Setup
        worker.instrument_configs[1] = InstrumentConfig(True, 5)
        worker._connection = MagicMock()

        nplc = "100"

        # Run test function
        worker.set_nplc(nplc)

        # Check the set NPLC command was set and the timeout was increased
        worker._connection.write.assert_any_call(f"NPLC {nplc}")
        assert worker._connection.timeout == 10000

    @pytest.mark.parametrize(
        "multimeter, invalid_value",
        [
            (Agilent3458A, "-1"),
            (Agilent3458A, "0"),
            (Agilent3458A, "2001"),
            (Agilent3458A, "999999"),
            (Agilent34401A, "-1"),
            (Agilent34401A, "0"),
            (Agilent34401A, "101"),
        ],
    )
    def test_set_nplc_invalid_value(
        self, multimeter: Type[Multimeter], invalid_value: str, worker: Worker
    ):
        """Test that set_nplc rejects invalid values"""

        worker.instrument_configs[1] = InstrumentConfig(
            True, 5, multimeter=multimeter()
        )
        worker._connection = MagicMock()
        worker.error = MagicMock()

        worker.set_nplc(invalid_value)

        worker.error.emit.assert_called_once_with(
            f"NPLC value {invalid_value} outside allowed range "
            f"{multimeter.nplc_min} - {multimeter.nplc_max}"
        )

    def test_validate_parameters(self, worker: Worker):
        """Test that validate_parameters allows through valid parameters"""

        worker.instrument_configs[1] = InstrumentConfig(True, 23)
        worker.interval = 1
        worker.writer = MagicMock()
        worker.nplc_input = MagicMock()

        assert worker.validate_parameters()

    @pytest.mark.parametrize(
        "address, interval, writer",
        [
            (0, 1, MagicMock()),
            (22, 0, MagicMock()),
            (22, -1, MagicMock()),
            (22, 1, None),
        ],
    )
    def test_validate_parameters_invalid(
        self, address: int, interval: int, writer: Any, worker: Worker
    ):
        """Test that validate_parameters rejects invalid parameters"""

        worker.instrument_configs[1] = InstrumentConfig(True, address)
        worker.interval = interval
        worker.writer = writer

        assert worker.validate_parameters() is False

    @patch("psc_datalogger.connection.pyvisa.ResourceManager")
    def test_init_connection(
        self,
        mock_resource_manager_init: MagicMock,
        worker: Worker,
    ):
        """Test that init_connection can create a valid connection"""

        mock_resource_manager = MagicMock()
        mock_resource_manager.list_resources = MagicMock(
            return_value=(
                "CONN1",
                "CONN2",
            )
        )
        mock_conn = MagicMock(spec=SerialInstrument)
        mock_resource_manager.open_resource = MagicMock(return_value=mock_conn)
        mock_resource_manager_init.return_value = mock_resource_manager

        worker._check_resource_is_prologix = MagicMock(return_value=True)

        worker.init_connection()

        assert worker._connection == mock_conn

    @patch("psc_datalogger.connection.pyvisa.ResourceManager")
    def test_init_connection_no_resources(
        self,
        mock_resource_manager_init: MagicMock,
        worker: Worker,
        caplog: pytest.LogCaptureFixture,
    ):
        """Test that init_connection emits expected error when there are no resources"""
        mock_resource_manager = MagicMock()
        mock_resource_manager.list_resources = MagicMock(return_value=())
        mock_resource_manager_init.return_value = mock_resource_manager

        with pytest.raises(PrologixNotFoundException):
            worker.init_connection()

        # Check the right log messages were created
        assert len(caplog.records) == 1
        assert caplog.records[0].message == "No Prologix controller found"

    @patch("psc_datalogger.connection.pyvisa.ResourceManager")
    def test_init_connection_invalid_resource(
        self,
        mock_resource_manager_init: MagicMock,
        worker: Worker,
        caplog: pytest.LogCaptureFixture,
    ):
        """Test that init_connection emits expected error when only invalid resources
        exist (i.e. none of them are a Prologix controller)"""

        mock_resource_manager = MagicMock()
        mock_resource_manager.list_resources = MagicMock(return_value=("CONN1",))
        mock_conn = MagicMock(spec=SerialInstrument)
        mock_resource_manager.open_resource = MagicMock(return_value=mock_conn)

        mock_resource_manager_init.return_value = mock_resource_manager

        worker._check_resource_is_prologix = MagicMock(return_value=False)

        with pytest.raises(PrologixNotFoundException):
            worker.init_connection()

        # Check the right log messages were created
        assert len(caplog.records) == 1
        assert caplog.records[0].message == "No Prologix controller found"

    def test_check_resource_is_prologix_with_prologix_resource(
        self,
        worker: Worker,
    ):
        """Test _check_resource_is_prologix returns True for a Prologix resource"""

        resource = MagicMock(spec=SerialInstrument)
        resource.query = MagicMock(return_value="Response to ++help!")

        assert worker._check_resource_is_prologix(resource) is True

    def test_check_resource_is_prologix_not_serial_instrument(
        self,
        worker: Worker,
    ):
        """Test _check_resource_is_prologix returns False if the resource is not a
        SerialInstrument"""

        resource = MagicMock(spec=RegisterBasedResource)

        assert worker._check_resource_is_prologix(resource) is False

    def test_check_resource_is_prologix_no_response(
        self,
        worker: Worker,
    ):
        """Test _check_resource_is_prologix returns False if there is no response to
        the query"""

        resource = MagicMock(spec=SerialInstrument)
        resource.query = MagicMock(return_value="")

        assert worker._check_resource_is_prologix(resource) is False

    def test_check_resource_is_prologix_query_timeout(
        self,
        worker: Worker,
    ):
        """Test _check_resource_is_prologix returns False if there is a timeout on the
        query"""

        resource = MagicMock(spec=SerialInstrument)
        resource.query = MagicMock(side_effect=pyvisa.VisaIOError(VI_ERROR_TMO))

        assert worker._check_resource_is_prologix(resource) is False

    def test_do_logging(self, worker: Worker):
        """Test do_logging sends a query_complete signal and passes data to the
        writer"""

        # Setup all mocks
        worker._connection = True  # type: ignore
        worker.writer = MagicMock()
        worker.query_complete = MagicMock()
        now = datetime.now()
        data = (now, "123", "456", "789")
        worker.query_instruments = MagicMock(return_value=data)

        # Call under test
        worker.do_logging()

        # Checks
        worker.query_complete.emit.assert_called_once_with(now)
        worker.writer.write.assert_called_once_with(
            timestamp=now, ins_1=data[1], ins_2=data[2], ins_3=data[3]
        )

    def test_do_logging_reports_error(self, worker: Worker):
        """Test do_logging sends an error signal with invalid data"""

        # Setup all mocks
        worker._connection = True  # type: ignore
        worker.writer = MagicMock()
        worker.error = MagicMock()
        now = datetime.now()
        data = (now, Worker.ERROR_STRING, "456", "789")
        worker.query_instruments = MagicMock(return_value=data)

        # Call under test
        worker.do_logging()

        # Checks
        worker.error.emit.assert_called_once()  # Don't bother checking string!
        worker.writer.write.assert_called_once_with(
            timestamp=now, ins_1=data[1], ins_2=data[2], ins_3=data[3]
        )

    @patch("psc_datalogger.multimeter.time")
    @pytest.mark.parametrize(
        "multimeter, expected_writes, expected_queries",
        [
            (
                Agilent34401A(),
                [
                    call("++addr 22"),
                    call("++auto 0"),
                    call("VOLT:DC:NPLCYCLES?"),
                    call("INIT"),
                    call("FETCH?"),
                ],
                [
                    call("++read eoi"),
                    call("++read eoi"),
                ],
            ),
            (
                Agilent3458A(),
                [call("++addr 22")],
                [
                    call("TRIG SGL"),
                ],
            ),
        ],
        ids=multimeter_naming,
    )
    def test_query_instrument_sends_expected_commands(
        self,
        mocked_time: MagicMock,
        worker: Worker,
        multimeter: Multimeter,
        expected_writes: List[_Call],
        expected_queries: List[_Call],
    ):
        """Test that the query_instrument function sends the right commands to the
        multimeter"""
        address = 22  # NOTE: Also in parameters
        worker.instrument_configs[1] = InstrumentConfig(
            True, address, multimeter=multimeter
        )

        mocked_write = MagicMock()
        mocked_query = MagicMock(return_value="1")
        worker._connection = MagicMock()
        worker._connection.write = mocked_write
        worker._connection.query = mocked_query

        # Call under test
        worker.query_instruments()

        # mocked_write.assert_called_once_with(f"++addr {address}")
        mocked_write.assert_has_calls(expected_writes, any_order=False)
        mocked_query.assert_has_calls(expected_queries, any_order=False)

    @patch("psc_datalogger.connection.datetime")
    @pytest.mark.parametrize(
        "multimeter, voltage_reading",
        [
            (Agilent34401A(), "\x009.089320482E+00"),
            (Agilent3458A(), "\x00 9.089320482E+00"),
        ],
        ids=multimeter_naming,
    )
    def test_query_instruments_voltage(
        self,
        mocked_datetime: MagicMock,
        worker: Worker,
        multimeter: Multimeter,
        voltage_reading: str,
    ):
        """Test querying the (mocked) hardware returns voltage"""
        # Set up mocks
        address = 22
        worker.instrument_configs[1] = InstrumentConfig(
            True, address, multimeter=multimeter
        )

        voltage_trimmed = "9.089320482E+00"  # The cleaned up string we expect
        mocked_write = MagicMock()
        mocked_query = MagicMock(return_value=voltage_reading)
        worker._connection = MagicMock()
        worker._connection.write = mocked_write
        worker._connection.query = mocked_query

        now = datetime.now()
        mocked_datetime.now.return_value = now

        # Call under test
        results = worker.query_instruments()

        # Assert results
        assert results[0] == now
        assert results[1] == voltage_trimmed
        assert results[2] == ""
        assert results[3] == ""

    @patch("psc_datalogger.connection.datetime")
    @pytest.mark.parametrize(
        "multimeter, query_returns",
        [
            # First "10" is for the NPLC query
            (Agilent34401A(), ["10", "\x001E-03"]),
            (Agilent3458A(), ["\x00 1E-03"]),
        ],
        ids=multimeter_naming,
    )
    def test_query_instruments_temperature(
        self,
        mocked_datetime: MagicMock,
        worker: Worker,
        multimeter: Multimeter,
        query_returns: str,
    ):
        """Test querying the (mocked) hardware returns the voltage converted to a
        temperature"""
        # Set up mocks
        address = 22
        worker.instrument_configs[1] = InstrumentConfig(
            True, address, multimeter=multimeter, convert_to_temp=True
        )
        temperature = "0.2"  # degrees Celcius, calculated from voltage_str
        mocked_write = MagicMock()
        mocked_query = MagicMock(side_effect=query_returns)
        worker._connection = MagicMock()
        worker._connection.write = mocked_write
        worker._connection.query = mocked_query

        now = datetime.now()
        mocked_datetime.now.return_value = now

        # Call under test
        results = worker.query_instruments()

        # Assert results
        assert results[0] == now
        assert results[1] == temperature
        assert results[2] == ""
        assert results[3] == ""

    @patch("psc_datalogger.connection.datetime")
    @pytest.mark.parametrize(
        "multimeter",
        [
            Agilent34401A(),
            Agilent3458A(),
        ],
        ids=multimeter_naming,
    )
    def test_query_instruments_timeout(
        self,
        mocked_datetime: MagicMock,
        worker: Worker,
        multimeter: Multimeter,
        caplog: LogCaptureFixture,
    ):
        """Test that a timeout returns an error string"""
        # Set up mocks
        address = 22
        worker.instrument_configs[1] = InstrumentConfig(
            True, address, multimeter=multimeter
        )
        mocked_write = MagicMock()
        mocked_query = MagicMock(side_effect=pyvisa.VisaIOError(VI_ERROR_TMO))
        worker._connection = MagicMock()
        worker._connection.write = mocked_write
        worker._connection.query = mocked_query

        now = datetime.now()
        mocked_datetime.now.return_value = now

        # Call under test
        results = worker.query_instruments()

        # Assert results
        assert results[0] == now
        assert results[1] == worker.ERROR_STRING
        assert results[2] == ""
        assert results[3] == ""

        assert "Exception reading from address " in caplog.text

    @patch("psc_datalogger.connection.datetime")
    @pytest.mark.parametrize(
        "multimeter",
        [
            Agilent34401A(),
            Agilent3458A(),
        ],
        ids=multimeter_naming,
    )
    def test_query_instruments_invalid_temperature_reading(
        self,
        mocked_datetime: MagicMock,
        worker: Worker,
        multimeter: Multimeter,
        caplog: LogCaptureFixture,
    ):
        """Test that an invalid voltage that cannot be converted to a temperature
        returns an error"""
        # Set up mocks
        address = 22
        worker.instrument_configs[1] = InstrumentConfig(
            True, address, multimeter=multimeter, convert_to_temp=True
        )
        voltage_str = "0.5"  # Value is too large to convert to temp
        mocked_write = MagicMock()
        mocked_query = MagicMock(return_value=voltage_str)
        worker._connection = MagicMock()
        worker._connection.write = mocked_write
        worker._connection.query = mocked_query

        now = datetime.now()
        mocked_datetime.now.return_value = now

        # Call under test
        results = worker.query_instruments()

        # Assert results
        assert results[0] == now
        assert results[1] == worker.ERROR_STRING
        assert results[2] == ""
        assert results[3] == ""

        assert f"Exception converting value {voltage_str} to temperature" in caplog.text

    def test_connection_check_no_connection(self, worker: Worker):
        """Test that when there is no connection the expected exception is raised"""
        with pytest.raises(ConnectionNotInitialized):
            worker._connection_check()



from datetime import datetime

from PyQt5.QtWidgets import QLabel, QStatusBar


class StatusBar(QStatusBar):
    """The StatusBar for the MainWindow application. Provides feedback to the user on
    the current state of the system"""

    def __init__(self, parent):
        super().__init__(parent)
        self.status_label = QLabel(text="")
        self.addWidget(self.status_label)

    def query_complete_callback(self, timestamp: datetime):
        # Remove microseconds, we don't need to display that much precision
        simple_timestamp = timestamp.isoformat(sep=" ", timespec="seconds")
        self.status_label.setText(
            f"Logging Running. Last data taken: {simple_timestamp}"
        )

    def init_complete_callback(self):
        self.status_label.setText("Serial Connection open")

    def error_callback(self, error: str):
        self.status_label.setText(f"ERROR: {error}")

    def logging_started(self):
        self.status_label.setText("Logging Running. No data collected yet.")

    def logging_stopped(self):
        self.status_label.setText("Logging Stopped.")

    def clear_status(self):
        self.status_label.setText("")



import logging
from typing import Callable, Optional, Type

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIntValidator
from PyQt5.QtWidgets import (
    QApplication,
    QButtonGroup,
    QFileDialog,
    QFrame,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QWidget,
)

from psc_datalogger.multimeter import Agilent3458A, Agilent34401A, Multimeter

from . import __version__
from .connection import ConnectionManager
from .statusbar import StatusBar


def application():
    app = QApplication([])

    window = DataloggerMainWindow()
    window.show()

    logging.debug("Launching application")
    app.exec()


class DataloggerMainWindow(QMainWindow):
    """The main window of the Datalogger application"""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle(f"PSC Datalogger {__version__}")

        self.connection_manager = ConnectionManager()

        self.create_widgets()

        self.connection_manager.set_status_bar(self.status_bar)
        self.connection_manager.set_nplc_input(self.nplc_input)
        self.connection_manager.start()

    def create_widgets(self) -> None:
        """Create all the widgets for this screen"""

        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)
        layout = QVBoxLayout()
        central_widget.setLayout(layout)

        # Widgets for each instrument
        self.instruments: list[AgilentWidgets] = []
        checked = True
        for i in range(3):
            # Use 1-index for GUI labels
            instrument = AgilentWidgets(i + 1, checked, self.handle_instrument_changed)
            self.instruments.append(instrument)
            layout.addWidget(instrument)
            checked = False

        # Widgets for the update interval
        interval_frame = QFrame()
        interval_label = QLabel("Update interval (s):")
        interval_layout = QHBoxLayout()
        interval_frame.setLayout(interval_layout)
        interval_layout.addWidget(interval_label)
        interval_input = QLineEdit()
        interval_input.setValidator(QIntValidator())
        interval_input.textEdited.connect(self.connection_manager.set_interval)
        interval_layout.addWidget(interval_input)
        layout.addWidget(interval_frame)

        logfile_widgets = LogFileWidgets(self.connection_manager.set_filepath)
        layout.addWidget(logfile_widgets)

        # Widgets for NPLC
        nplc_frame = QFrame()
        nplc_label = QLabel("NPLC:")
        nplc_layout = QHBoxLayout()
        nplc_frame.setLayout(nplc_layout)
        nplc_layout.addWidget(nplc_label)
        self.nplc_input = QLineEdit()
        self.nplc_input.setText("50")
        self.nplc_input.setValidator(QIntValidator())
        self.nplc_input.textEdited.connect(self.connection_manager.set_nplc)
        nplc_layout.addWidget(self.nplc_input)
        layout.addWidget(nplc_frame)

        # Start/Stop logging buttons
        self.start_text = "Start Logging"
        self.stop_text = "Stop Logging"
        self.start_stop_button = QPushButton(self.start_text)
        self.start_stop_button.setCheckable(True)
        self.start_stop_button.toggled.connect(self.handle_start_stop)
        layout.addWidget(self.start_stop_button)

        self.status_bar = StatusBar(self)
        self.setStatusBar(self.status_bar)

    def handle_start_stop(self, checked):
        """Handles starting and stopping logging"""
        logging.debug(f"Start/Stop button is {checked}")
        if checked:
            # Button is pressed, do start actions
            self.start_stop_button.setText(self.stop_text)
            started = self.connection_manager.start_logging()
            if not started:
                error_dialog = QMessageBox(self)
                error_dialog.setIcon(QMessageBox.Icon.Critical)
                error_dialog.setModal(True)
                error_dialog.setWindowTitle("Failed to start logging")
                error_dialog.setText("Unable to start logging data")
                error_dialog.setInformativeText(
                    "Check at least one address, update interval, and logfile is set"
                )
                error_dialog.exec()
                self.start_stop_button.setChecked(False)
        else:
            # Button is unpressed, do stop actions
            self.start_stop_button.setText(self.start_text)
            self.connection_manager.stop_logging()

    def handle_instrument_changed(self):
        """Handle when any of the instrument configurations change"""
        for instrument in self.instruments:
            try:
                self.connection_manager.set_instrument(
                    instrument.instrument_number,
                    instrument.isChecked(),
                    instrument.get_address(),
                    instrument.get_temperature_checked(),
                    instrument.get_multimeter(),
                )
            except ValueError:
                # Expected when first activating an instrument; the Address field
                # will be empty
                logging.info(
                    "Exception when initializing instrument "
                    f"{instrument.instrument_number}",
                    exc_info=True,
                )


class LogFileWidgets(QFrame):
    """Contains widgets related to selecting a log file"""

    def __init__(self, callback: Callable[[str], None]) -> None:
        super().__init__()

        self.create_widgets()

        self.filename_callback = callback

        self.filename: Optional[str] = None

    def create_widgets(self) -> None:
        file_layout = QHBoxLayout()
        # This is actually a pop-up dialog, not an inline widget
        self.dialog = QFileDialog(None, "Select output logfile")
        self.dialog.setFileMode(QFileDialog.FileMode.AnyFile)
        self.dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        self.dialog.setNameFilter("CSV (*.csv)")
        self.dialog.setDefaultSuffix(".csv")
        self.dialog.fileSelected.connect(self.file_selected)

        file_button = QPushButton("Select Logfile...")
        # Suppress error that `exec` doesn't technically match signature of a PYQT_SLOT
        # - it doesn't affect anything, the return value is discarded
        file_button.clicked.connect(self.dialog.exec)  # type: ignore
        file_layout.addWidget(file_button)

        self.file_readback = QLineEdit()
        self.file_readback.setDisabled(True)
        file_layout.addWidget(self.file_readback)

        self.setLayout(file_layout)

    def file_selected(self, new_path: str) -> None:
        """Handle a new filepath being set"""
        self.filename = new_path
        self.file_readback.setText(new_path)
        self.filename_callback(new_path)


class AgilentWidgets(QGroupBox):
    """Contains widgets to configure a single instrument"""

    def __init__(
        self, instrument_number: int, checked: bool, instrument_changed: Callable
    ) -> None:
        """
        Args:
          instrument_number: The number to use to identify this instance
          checked: True if the widgets should be enabled by default
          instrument_changed: Callback to be called whenever any configuration changes
        """
        super().__init__(f"Instrument {instrument_number}")
        self.instrument_number = instrument_number

        self.setCheckable(True)
        self.setChecked(checked)
        self.toggled.connect(instrument_changed)

        self.create_widgets(instrument_changed)

    def create_widgets(self, instrument_changed: Callable) -> None:
        address_label = QLabel("GPIB Address")
        self.address_input_box = QLineEdit()
        int_only = QIntValidator()
        int_only.setRange(0, 30)
        self.address_input_box.setValidator(int_only)
        self.address_input_box.editingFinished.connect(instrument_changed)
        self.voltage_radiobutton = QRadioButton("Voltage")
        self.voltage_radiobutton.setChecked(True)
        self.voltage_radiobutton.toggled.connect(instrument_changed)
        self.temperature_radiobutton = QRadioButton("Temperature")
        # Don't need to .connect() the second button as changing the first one
        # triggers a callback that will check both buttons

        grid_layout = QGridLayout(self)

        first_row_layout = QHBoxLayout()

        self.A3458_radiobutton = QRadioButton("3458A")
        self.A3458_radiobutton.setChecked(True)
        self.A3458_radiobutton.toggled.connect(instrument_changed)
        self.A34401_radiobutton = QRadioButton("34401A")

        device_groupbox = QButtonGroup(self)
        device_groupbox.addButton(self.A3458_radiobutton)
        device_groupbox.addButton(self.A34401_radiobutton)

        first_row_layout.addWidget(self.A3458_radiobutton)
        first_row_layout.addWidget(self.A34401_radiobutton)

        grid_layout.addLayout(first_row_layout, 0, 0, Qt.AlignmentFlag.AlignCenter)

        second_row_layout = QHBoxLayout()

        measurement_type_groupbox = QButtonGroup(self)
        measurement_type_groupbox.addButton(self.voltage_radiobutton)
        measurement_type_groupbox.addButton(self.temperature_radiobutton)

        second_row_layout.addWidget(address_label)
        second_row_layout.addWidget(self.address_input_box)
        second_row_layout.addWidget(self.voltage_radiobutton)
        second_row_layout.addWidget(self.temperature_radiobutton)

        grid_layout.addLayout(second_row_layout, 1, 0, Qt.AlignmentFlag.AlignCenter)

        self.setLayout(grid_layout)

    def get_address(self) -> str:
        return self.address_input_box.text()

    def get_temperature_checked(self) -> bool:
        """Returns True if this instrument is configured to read temperature instead of
        voltage"""
        return self.temperature_radiobutton.isChecked()

    def get_multimeter(self) -> Type[Multimeter]:
        return Agilent3458A if self.A3458_radiobutton.isChecked() else Agilent34401A



import logging
import time
from collections import OrderedDict
from csv import DictWriter
from dataclasses import dataclass
from datetime import datetime
from threading import Event, RLock
from typing import List, Optional, TextIO, Tuple, Type, cast

import pyvisa
from PyQt5.QtCore import QObject, QThread, pyqtSignal
from PyQt5.QtWidgets import QLineEdit
from pyvisa.resources import Resource, SerialInstrument

from .multimeter import Agilent3458A, InvalidNplcException, Multimeter
from .statusbar import StatusBar
from .temperature.converter import volts_to_celcius


class ConnectionManager:
    """Manage the connection to the instruments, which is handled on a separate
    thread."""

    def __init__(self):
        self.thread = QThread()
        # Use "signal" as "event" is used by PyQt
        self.logging_signal = Event()
        self._worker = Worker(self.logging_signal)
        self._worker.moveToThread(self.thread)
        self.thread.started.connect(self._worker.run)

    def start(self):
        """Begin the worker thread"""
        logging.debug("Starting worker thread")
        self.thread.start()

    def set_status_bar(self, status_bar: StatusBar):
        """Set up the status bar to display relevant information"""
        self.status_bar = status_bar
        self._worker.query_complete.connect(status_bar.query_complete_callback)
        self._worker.init_complete.connect(status_bar.init_complete_callback)
        self._worker.error.connect(status_bar.error_callback)
        self._worker.clear_status.connect(status_bar.clear_status)

    def set_nplc_input(self, nplc_input: QLineEdit):
        """Save the GUI component that houses the NPLC value"""
        self._worker.nplc_input = nplc_input

    def set_interval(self, interval: str):
        """Set the update interval for the logging thread"""
        if interval != "":  # Empty string passed if the textbox is cleared
            self._worker.interval = float(interval)

    def set_filepath(self, filepath):
        """Set the filepath for the logging thread"""
        self._worker.set_filepath(filepath)

    def set_instrument(
        self,
        instrument_number: int,
        enabled: bool,
        gpib_address: str,
        measure_temp: bool,
        multimeter_type: Type[Multimeter],
    ) -> None:
        """Configure the given instrument number with the provided parameters"""
        try:
            self._worker.set_instrument(
                instrument_number, enabled, gpib_address, measure_temp, multimeter_type
            )
        except ConnectionNotInitialized:
            logging.exception(
                f"Could not set instrument {instrument_number}, parmameters: "
                f"enabled {enabled} address {gpib_address} measure_temp {measure_temp}"
            )

    def set_nplc(self, nplc: str) -> None:
        """Configure the NPLC value for all instruments"""
        try:
            self._worker.set_nplc(nplc)
        except ConnectionNotInitialized:
            logging.exception(f"Could not set NPLC to {nplc}")

    def start_logging(self) -> bool:
        """Start the logging process in the background thread.
        Returns True if logging successfully started, else False"""
        if self._worker.validate_parameters():
            self.logging_signal.set()
            self.status_bar.logging_started()
            return True
        else:
            return False

    def stop_logging(self) -> None:
        """Stop the logging process in the background thread"""
        self.logging_signal.clear()
        self.status_bar.logging_stopped()


@dataclass
class InstrumentConfig:
    """Contains the configuration for a single instrument"""

    # Whether this instrument is marked enabled in the GUI
    enabled: bool = False
    # GPIB address of instrument
    address: int = -1
    # Indicate whether the voltage read should be converted into a temperature
    convert_to_temp: bool = False
    # The multimeter in use. Default to 3458A, matching the GUI's default
    multimeter: Multimeter = Agilent3458A()


class DataWriter(DictWriter):
    """Class that handles writing data to a file"""

    file: TextIO

    csv_fieldnames = [
        "timestamp",
        "instrument 1",
        "instrument 2",
        "instrument 3",
    ]

    def __init__(self, filepath: str):
        self.file = open(filepath, "w", newline="")

        super().__init__(self.file, fieldnames=self.csv_fieldnames, dialect="excel")

        self.writeheader()
        self.file.flush()

    def close(self) -> None:
        self.file.close()

    def write(self, timestamp: datetime, ins_1: str, ins_2: str, ins_3: str):
        """Write the given data to the file"""
        len_written = self.writerow(
            {
                self.csv_fieldnames[0]: str(timestamp),
                self.csv_fieldnames[1]: ins_1,
                self.csv_fieldnames[2]: ins_2,
                self.csv_fieldnames[3]: ins_3,
            }
        )
        logging.debug(f"DictWriter wrote {len_written} bytes")

        self.file.flush()
        logging.debug("File flushed")


class PrologixNotFoundException(Exception):
    """Exception thrown when the Prologix controller is not found"""


class ConnectionNotInitialized(Exception):
    """Exception thrown when attempting to access the serial connection when
    it has not yet been initialized."""


class Worker(QObject):
    """Class that does the serial connection to the instruments

    NOTE: This expects to be run in a separate QThread from the main GUI"""

    # Signal that initialization has completed
    init_complete = pyqtSignal()

    # Signal that we have queried all the instruments.
    # Parameter is the timestamp this occurred at.
    query_complete = pyqtSignal(datetime)

    # Signal that an error has occurred. Parameter is the error message.
    error = pyqtSignal(str)

    # Signal that will clear the status bar of any text.
    # Used to clear it after errors occur.
    clear_status = pyqtSignal()

    # The update interval that readings should be taken at
    interval: float = 0  # seconds

    writer: Optional[DataWriter] = None

    # Keep track of the configuration of each of the 3 possible devices
    # Ordered dict to ensure that we always read instruments in order when iterating
    instrument_configs = OrderedDict(
        {1: InstrumentConfig(), 2: InstrumentConfig(), 3: InstrumentConfig()}
    )

    # The connection to the Prologix controller. Do not directly access, use the
    # _connection_* wrapper functions.
    _connection: SerialInstrument

    # Constant to mark a measurement could not be taken. Also written to results file.
    ERROR_STRING = "#ERROR"

    # The GUI widget that contains the NPLC value
    nplc_input: QLineEdit

    def __init__(self, logging_signal: Event):
        """
        Create a Worker instance.

        Args:
            logging_signal: An Event object that, when set, means logging should run
        """
        super().__init__()
        self.logging_signal = logging_signal

        # This lock protects both the file and the serial connection resources
        # Recursive to allow more defensive programming
        self.lock = RLock()

        self.running = True

    def set_filepath(self, filepath: str):
        """Create the writer for the given filepath. Closes any existing
        writer/filehandle."""
        logging.info(f"Setting filepath to {filepath}")
        with self.lock:
            if self.writer:
                self.writer.close()

            self.writer = DataWriter(filepath)

    def set_instrument(
        self,
        instrument_number: int,
        enabled: bool,
        gpib_address: str,
        measure_temp: bool,
        multimeter_type: Type[Multimeter],
    ) -> None:
        """Configure the given instrument number with the provided parameters"""
        assert (
            1 <= instrument_number <= 3
        ), f"Invalid instrument number {instrument_number}"

        try:
            address = int(gpib_address)
        except ValueError:
            # May not be an address set in the GUI; this happens when the "enable"
            # tickbox is first pressed and the address field is likely empty.
            logging.info(f"gpib address '{gpib_address}' invalid")
            address = -1

        logging.info(
            f"Configuring instrument {instrument_number}; Enabled {enabled},"
            f"Address {gpib_address}, measure temp {measure_temp}"
        )
        self.instrument_configs[instrument_number] = InstrumentConfig(
            enabled, address, measure_temp, multimeter_type()
        )

        self._init_instrument(self.instrument_configs[instrument_number])

    def _init_instrument(self, instrument: InstrumentConfig) -> None:
        """Initialize the given instrument"""
        logging.debug(f"Initializing instrument at address {instrument.address}")

        if not instrument.enabled:
            logging.debug(
                f"Skipping _init_instrument for instrument {instrument.address}"
            )
            return

        gpib_address = instrument.address

        if gpib_address <= 0:
            raise ValueError(
                f"_init_instrument called with invalid address '{gpib_address}'"
            )

        multimeter = instrument.multimeter

        with self.lock:
            self._set_prologix_address(gpib_address)

            multimeter.initialize(self._connection)

            self._set_nplc(instrument, int(self.nplc_input.text()))

            # Read all data remaining in the buffer; it is possible for
            # samples to be taken while initializing the multimeters.
            # (Mostly an issue with the 3458A but doesn't hurt for other types)
            while self._connection_bytes_in_buffer():
                try:
                    self._connection_read()
                except pyvisa.VisaIOError:
                    logging.debug(f"Instrument {gpib_address} data buffer emptied")

        logging.debug(f"Instrument initialized at address {instrument.address}")

    def set_nplc(self, nplc: str) -> None:
        """Set the NPLC for all configured instruments"""

        try:
            int_plc = int(nplc)
        except ValueError:
            # This happens if the input text box is empty
            return

        try:
            for instrument in self.instrument_configs.values():
                if instrument.enabled:
                    self._set_nplc(instrument, int_plc)
        except InvalidNplcException as e:
            self.error.emit(
                f"NPLC value {int_plc} outside allowed range "
                f"{e.min_allowed} - {e.max_allowed}"
            )
            return

        # Assuming all devices accepted the NPLC, clear any errors
        self.clear_status.emit()

    def _set_nplc(self, instrument: InstrumentConfig, nplc: int) -> None:
        """Set the NPLC for the given instrument"""
        with self.lock:
            self._set_prologix_address(instrument.address)

            multimeter = instrument.multimeter

            multimeter.set_nplc(self._connection, nplc)

            # The number of samples is tied to the electrical frequency.
            # i.e. an NPLC of 50 will take 1 second, as our mains runs at 50Hz.
            # So we need to ensure our timeout is more-than-enough to cover it
            # However we want to ensure we always have at least a 5 second timeout
            # as even the smallest NPLC values seem to take a few seconds to complete
            timeout = max(5000, nplc * 100)
            self._set_connection_timeout(timeout)  # ms

    def _set_prologix_address(self, gpib_address: int) -> None:
        """Configure the Prologix to point to the given GPIB address"""
        with self.lock:
            self._connection_write(f"++addr {gpib_address}")
            # Prologix seems to need a moment to process previous command
            time.sleep(0.1)

    def validate_parameters(self) -> bool:
        """Returns True if all required parameters are set, otherwise False"""

        if all(x.address <= 0 for x in self.instrument_configs.values()):
            logging.warning("No GPIB addresses set for any instrument")
            return False

        if self.interval <= 0:
            logging.warning("No update interval set")
            return False

        if self.writer is None:
            logging.warning("No logfile selected")
            return False

        if self.nplc_input.text() == "":
            logging.warning("No NPLC value set")
            return False

        logging.info("Parameters are valid")
        return True

    def run(self):
        """Main work function of this class. Initializes a connection then continually
        queries the instruments for data, and logs it"""

        while True:
            try:
                self.init_connection()
                self.init_complete.emit()
                break
            except PrologixNotFoundException:
                errmsg = "Prologix controller not found. Trying again in 5 seconds"
                logging.exception(errmsg)
                self.error.emit(errmsg)
                time.sleep(5)

        while self.running:
            self.logging_signal.wait()

            logging.debug("Starting log loop")

            try:
                self.do_logging()
            except Exception:
                # Ignore it and continue working
                logging.exception("Unexpected exception while logging data")
                pass

            # Sleeping like this will cause minor drift over time, equal to how long
            # reading from all instruments takes. This becomes a major problem if
            # timeouts occur as they invoke a 3-second delay per timeout.
            time.sleep(self.interval)
            logging.debug("Log sleep finished")

    def init_connection(self):
        """Initialize the connection to the Prologix device"""
        with self.lock:
            rm = pyvisa.ResourceManager()

            resources = rm.list_resources()

            logging.info(f"Resources available: {resources}")

            # Find the connection that is the Prologix controller
            # Reversed as it's more common for it to be the last resource in the list
            for resource in reversed(resources):
                conn = rm.open_resource(resource)
                if self._check_resource_is_prologix(conn):
                    logging.info(f"Found Prologix controller at {resource}")
                    break
            else:
                rm.close()
                errmsg = "No Prologix controller found"
                logging.error(errmsg)
                raise PrologixNotFoundException(errmsg)

            # The open_resource function returns a very generic type
            self._connection = cast(SerialInstrument, conn)

            logging.info("Connection initialized")

    def _check_resource_is_prologix(self, resource: Resource) -> bool:
        """Returns True if the given Resource is a Prologix controller, otherwise
        False"""
        try:
            assert isinstance(resource, SerialInstrument)
            # We assume that only a Prologix controller will respond to ++help
            help_str = resource.query("++help")
            if help_str:
                return True
        except pyvisa.VisaIOError:
            # Timeout; probably not the right device!
            logging.debug(f"No response to ++help for resource {resource}")
            return False
        except AssertionError:
            # Not a SerialInstrument, definitely not the correct resource
            logging.debug(f"Resource {resource} is not a SerialInstrument")
            return False

        return False

    def do_logging(self):
        """Take one set of readings and write them to file"""

        with self.lock:
            self._connection_check()
            assert self.writer is not None

            results = self.query_instruments()

            if any(x == self.ERROR_STRING for x in results):
                simple_timestamp = results[0].isoformat(sep=" ", timespec="seconds")
                logging.error("Unable to read data")
                self.error.emit(f"Unable to read data {simple_timestamp}")
            else:
                self.query_complete.emit(results[0])

            logging.info(
                f"Data read: {str(results[0])} {results[1]} {results[2]} {results[3]}"
            )

            self.writer.write(
                timestamp=results[0],
                ins_1=results[1],
                ins_2=results[2],
                ins_3=results[3],
            )

    def query_instruments(self) -> Tuple[datetime, str, str, str]:
        """Query the instruments and return the timestamp followed by three instrument
        readings."""

        with self.lock:
            measurement_time = datetime.now()

            measurements: List[str] = []
            for i in self.instrument_configs.values():
                logging.debug(f"Querying instrument {i.address}")
                if i.address <= 0:
                    # No address, add empty entry
                    measurements.append("")
                    continue

                try:
                    # Configure Prologix to talk to the current device
                    self._connection_write(f"++addr {i.address}")

                    logging.debug(f"Triggering instrument {i.address}")

                    multimeter = i.multimeter

                    # Request a single measurement
                    val = multimeter.take_reading(self._connection)

                    logging.debug(f"Address {i.address} Value {val}")

                except Exception:
                    # Issue reading from this instrument. Mark an error but continue
                    # processing other instruments
                    logging.exception(f"Exception reading from address {i.address}")
                    val = self.ERROR_STRING
                else:
                    try:
                        if i.convert_to_temp:
                            val = str(volts_to_celcius(val))
                    except AssertionError:
                        # Issue converting value to temperature. Mark an error but
                        # continue processing other instruments
                        logging.exception(
                            f"Exception converting value {val} to "
                            f"temperature from address {i.address}"
                        )
                        val = self.ERROR_STRING

                measurements.append(val)

            assert len(measurements) == 3
            measurement_tuple = cast(Tuple[str, str, str], tuple(measurements))
            return measurement_time, *measurement_tuple

    def _exit(self) -> None:
        """Cleanly stop the run() method to terminate all processing.
        This method is only used in testing!"""
        if self.writer:
            self.writer.close()
        # Send relevant flags to allow run() to terminate
        # Note it will do 1 more iteration of the loop, including the sleep
        self.running = False
        self.logging_signal.set()

    def _connection_check(self) -> None:
        """Check whether connection is initialized.

        Mostly required to guard against potential checks before the Prologix
        controller has connected (or maybe there isn't even one plugged in yet!)"""
        if not hasattr(self, "_connection"):
            logging.warning(
                "Attempted to write to connection when it was not initialized"
            )
            raise ConnectionNotInitialized()

    def _connection_write(self, command: str) -> None:
        """Write the given command to the connection"""
        self._connection_check()
        self._connection.write(command)

    def _connection_bytes_in_buffer(self) -> int:
        """Access connection.bytes_in_buffer"""
        self._connection_check()
        return self._connection.bytes_in_buffer

    def _connection_read(self) -> str:
        """Read data from the connection"""
        self._connection_check()
        return self._connection.read()

    def _set_connection_timeout(self, timeout: int) -> None:
        """Set timeout on the connection.

        Timeout is in milliseconds."""
        self._connection_check()
        self._connection.timeout = timeout

    def _connection_query(self, command: str) -> str:
        """Write the given command to the connection and read the result"""
        self._connection_check()
        return self._connection.query(command)



import logging
import subprocess
import sys

import pytest

from psc_datalogger import __version__
from psc_datalogger.__main__ import parse_args


def test_cli_version():
    cmd = [sys.executable, "-m", "psc_datalogger", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_cli_script_entrypoint_version():
    """Check that the entrypoint defined in [project.scripts] inside
    pyproject.toml works"""
    cmd = ["psc-datalogger", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


@pytest.mark.parametrize(
    "level_str, level_const",
    [
        ("DEBUG", logging.DEBUG),
        ("INFO", logging.INFO),
        ("WARNING", logging.WARNING),
        ("ERROR", logging.ERROR),
        ("CRITICAL", logging.CRITICAL),
    ],
)
def test_parse_log_level(level_str, level_const):
    args = parse_args(["-l", level_str])
    assert args.log_level == level_const



import logging
import time
from abc import ABC, abstractmethod

from pyvisa.resources import SerialInstrument


class InvalidNplcException(Exception):
    """Exception thrown when user attempts to input an NPLC value that is
    not valid for one-or-more multimeters"""

    min_allowed: float
    max_allowed: float

    def __init__(self, min_allowed: float, max_allowed: float, *args):
        self.min_allowed = min_allowed
        self.max_allowed = max_allowed
        super().__init__(args)


class Multimeter(ABC):
    """Generic type of Multimeters. Provides interface for possible operations
    on different controllers"""

    # The minimum and maximum valid NPLC values.
    nplc_min = None
    nplc_max = None

    def __init__(self):
        if self.nplc_min is None or self.nplc_max is None:
            raise NotImplementedError(
                "Subclasses must define nplc_min and nplc_max attributes"
            )

    @abstractmethod
    def initialize(self, connection: SerialInstrument):
        """Perform the commands to  initialize (or reset) the multimeter to a
        known state"""

    @abstractmethod
    def set_nplc(self, connection: SerialInstrument, nplc: int):
        """Perform the commands to set the given nplc. Raises InvalidNplcException if
        nplc is invalid."""

    @abstractmethod
    def take_reading(self, connection: SerialInstrument) -> str:
        """Perform the commands to return a single voltage reading"""


class Agilent3458A(Multimeter):
    """Implements the required commands for an Agilent 3458A multimeter"""

    nplc_min = 1
    nplc_max = 2000

    def initialize(self, connection: SerialInstrument):
        self._ensure_prologix_settings(connection)
        connection.write("PRESET NORM")  # Set a variety of defaults
        connection.write("BEEP 0")  # Disable annoying beeps

        # This means the instrument will stop collecting measurements, thus
        # not filling its internal memory buffer. Later we will send single
        # trigger events and immediately read it, thus keeping the buffer
        # empty so we avoid reading stale results
        connection.write("TRIG HOLD")

        logging.info("Initialized Agilent3458A")

    def set_nplc(self, connection: SerialInstrument, nplc: int):
        self._ensure_prologix_settings(connection)
        if not self.nplc_min <= nplc <= self.nplc_max:
            raise InvalidNplcException(self.nplc_min, self.nplc_max)

        connection.write(f"NPLC {nplc}")

    def take_reading(self, connection: SerialInstrument) -> str:
        # Send single trigger command to the multimeter, and return the resultant value
        raw = connection.query("TRIG SGL")
        return self._clean_string(raw)

    def _clean_string(self, volts: str) -> str:
        """Clean up the returned string by removing extraneous characters"""
        # Value format is e.g. " 9.089320482E+00\r\n"
        # Occasionally there are also leading NULL bytes.
        return volts.replace("\x00", "").strip(" \r\n")

    def _ensure_prologix_settings(self, connection: SerialInstrument):
        connection.write("++auto 1")  # Enables use of "query"
        connection.read_termination = "\r\n"


class Agilent34401A(Multimeter):
    """Implements the required commands for an Agilent 34401A multimeter"""

    nplc_min = 0.02
    nplc_max = 100

    def initialize(self, connection: SerialInstrument):
        self._ensure_prologix_settings(connection)
        connection.write("*RST")  # Reset the multimeter to its power-on configuration.
        # Configure for voltage reading with range of 10V, resolution of 0.003
        connection.write("CONFigure:VOLTage:DC 10, 0.003")
        connection.write("SYSTEM:BEEPER:STATE OFF")  # Disable system beeps

        logging.info("Initialized Agilent34401A")

    def set_nplc(self, connection: SerialInstrument, nplc: int):
        self._ensure_prologix_settings(connection)

        if not self.nplc_min <= nplc <= self.nplc_max:
            raise InvalidNplcException(self.nplc_min, self.nplc_max)

        connection.write(f"VOLT:DC:NPLCYCLES {nplc}")

    def take_reading(self, connection: SerialInstrument) -> str:
        self._ensure_prologix_settings(connection)

        # Get the currently configured NPLC setting, to allow us to estimate how
        # long the subsequent voltage reading will take
        connection.write("VOLT:DC:NPLCYCLES?")
        nplc = connection.query("++read eoi")

        nplc_float = float(self._clean_string(nplc))

        assert self.nplc_min <= nplc_float <= self.nplc_max, "Invalid NPLC read"

        # Triggers a single reading of the multimeter, which will save the data
        # into internal memory. This can take several seconds, based on NPLC setting.
        connection.write("INIT")

        # Estimate how long to wait. An NPLC of 50 should take 1 second (matches
        # electrical frequency). Add 1 for some padding.
        wait_time = (nplc_float / 50) + 1
        time.sleep(wait_time)

        # Data should now be in internal memory; tell the multimeter to move it
        # to output buffer, then tell the Prologix controller to read it.
        connection.write("FETCH?")
        volts = connection.query("++read eoi")

        return self._clean_string(volts)

    def _clean_string(self, volts: str) -> str:
        """Clean up the returned string by removing extraneous characters"""
        return volts.replace("\x00", "")

    def _ensure_prologix_settings(self, connection: SerialInstrument):
        # "auto 1" causes "error -420 Query Unterminated" errors when doing write
        # commands that don't return any data, so must do manual explicit write and read
        connection.write("++auto 0")
        connection.read_termination = "\n"
